@@ -15,19 +15,19 @@ type Node struct {
 
 	NodeType NodeType
 	// Only one of the below may be non-null.
-	Datasource          *Datasource
-	Distinct            *Distinct
-	Filter              *Filter
-	GroupBy             *GroupBy
-	LookupJoin          *LookupJoin
-	StreamJoin          *StreamJoin
-	Map                 *Map
-	OrderBy             *OrderBy
-	Requalifier         *Requalifier
-	TableValuedFunction *TableValuedFunction
-	Unnest              *Unnest
-	Limit               *Limit
-	InMemoryRecords     *InMemoryRecords
+	Datasource              *Datasource
+	Distinct                *Distinct
+	Filter                  *Filter
+	GroupBy                 *GroupBy
+	LookupJoin              *LookupJoin
+	StreamJoin              *StreamJoin
+	Map                     *Map
+	Requalifier             *Requalifier
+	TableValuedFunction     *TableValuedFunction
+	Unnest                  *Unnest
+	InMemoryRecords         *InMemoryRecords
+	OuterJoin               *OuterJoin
+	OrderSensitiveTransform *OrderSensitiveTransform
 }
 
 type Schema struct {
@@ -71,12 +71,14 @@ const (
 	NodeTypeLookupJoin
 	NodeTypeStreamJoin
 	NodeTypeMap
-	NodeTypeOrderBy
+	_
 	NodeTypeRequalifier
 	NodeTypeTableValuedFunction
 	NodeTypeUnnest
-	NodeTypeLimit
+	_
 	NodeTypeInMemoryRecords
+	NodeTypeOuterJoin
+	NodeTypeOrderSensitiveTransform
 )
 
 func (t NodeType) String() string {
@@ -95,18 +97,18 @@ func (t NodeType) String() string {
 		return "stream_join"
 	case NodeTypeMap:
 		return "map"
-	case NodeTypeOrderBy:
-		return "order_by"
 	case NodeTypeRequalifier:
 		return "requalifier"
 	case NodeTypeTableValuedFunction:
 		return "table_valued_function"
 	case NodeTypeUnnest:
 		return "unnest"
-	case NodeTypeLimit:
-		return "limit"
 	case NodeTypeInMemoryRecords:
 		return "in_memory_records"
+	case NodeTypeOuterJoin:
+		return "outer_join"
+	case NodeTypeOrderSensitiveTransform:
+		return "order_sensitive_transform"
 	}
 	return "unknown"
 }
@@ -198,12 +200,6 @@ type Map struct {
 	Expressions []Expression
 }
 
-type OrderBy struct {
-	Source               Node
-	Key                  []Expression
-	DirectionMultipliers []int
-}
-
 type Requalifier struct {
 	Source    Node
 	Qualifier string
@@ -264,15 +260,23 @@ type Unnest struct {
 	Field  string
 }
 
-type Limit struct {
-	Source Node
-	Limit  Expression
-}
-
 type InMemoryRecords struct {
 	Records []execution.Record
 }
 
+type OuterJoin struct {
+	Left, Right       Node
+	LeftKey, RightKey []Expression
+	IsLeft, IsRight   bool // Full Outer Join will have both true.
+}
+
+type OrderSensitiveTransform struct {
+	Source                      Node
+	OrderByKey                  []Expression
+	OrderByDirectionMultipliers []int
+	Limit                       *Expression
+}
+
 func (node *Node) Materialize(ctx context.Context, env Environment) (execution.Node, error) {
 	switch node.NodeType {
 	case NodeTypeDatasource:
@@ -393,20 +397,6 @@ func (node *Node) Materialize(ctx context.Context, env Environment) (execution.N
 		}
 
 		return nodes.NewMap(source, expressions), nil
-	case NodeTypeOrderBy:
-		source, err := node.OrderBy.Source.Materialize(ctx, env)
-		if err != nil {
-			return nil, fmt.Errorf("couldn't materialize order by source: %w", err)
-		}
-		keyExprs := make([]execution.Expression, len(node.OrderBy.Key))
-		for i := range node.OrderBy.Key {
-			expr, err := node.OrderBy.Key[i].Materialize(ctx, env.WithRecordSchema(node.OrderBy.Source.Schema))
-			if err != nil {
-				return nil, fmt.Errorf("couldn't materialize order by key with index %d: %w", i, err)
-			}
-			keyExprs[i] = expr
-		}
-		return nodes.NewBatchOrderBy(source, keyExprs, node.OrderBy.DirectionMultipliers), nil
 	case NodeTypeRequalifier:
 		return node.Requalifier.Source.Materialize(ctx, env)
 
@@ -433,19 +423,70 @@ func (node *Node) Materialize(ctx context.Context, env Environment) (execution.N
 
 		return nodes.NewUnnest(source, index), nil
 
-	case NodeTypeLimit:
-		source, err := node.Limit.Source.Materialize(ctx, env)
-		if err != nil {
-			return nil, fmt.Errorf("couldn't materialize limit source: %w", err)
-		}
-		limit, err := node.Limit.Limit.Materialize(ctx, env)
-		if err != nil {
-			return nil, fmt.Errorf("couldn't materialize limit expression: %w", err)
-		}
-		return nodes.NewLimit(source, limit), nil
-
 	case NodeTypeInMemoryRecords:
 		return nodes.NewInMemoryRecords(node.InMemoryRecords.Records), nil
+
+	case NodeTypeOuterJoin:
+		left, err := node.OuterJoin.Left.Materialize(ctx, env)
+		if err != nil {
+			return nil, fmt.Errorf("couldn't materialize left join source: %w", err)
+		}
+		right, err := node.OuterJoin.Right.Materialize(ctx, env)
+		if err != nil {
+			return nil, fmt.Errorf("couldn't materialize right join source: %w", err)
+		}
+
+		leftKeyExprs := make([]execution.Expression, len(node.OuterJoin.LeftKey))
+		for i := range node.OuterJoin.LeftKey {
+			expr, err := node.OuterJoin.LeftKey[i].Materialize(ctx, env.WithRecordSchema(node.OuterJoin.Left.Schema))
+			if err != nil {
+				return nil, fmt.Errorf("couldn't materialize stream join left key expression with index %d: %w", i, err)
+			}
+			leftKeyExprs[i] = expr
+		}
+		rightKeyExprs := make([]execution.Expression, len(node.OuterJoin.RightKey))
+		for i := range node.OuterJoin.RightKey {
+			expr, err := node.OuterJoin.RightKey[i].Materialize(ctx, env.WithRecordSchema(node.OuterJoin.Right.Schema))
+			if err != nil {
+				return nil, fmt.Errorf("couldn't materialize stream join right key expression with index %d: %w", i, err)
+			}
+			rightKeyExprs[i] = expr
+		}
+
+		return nodes.NewOuterJoin(left, right, len(node.OuterJoin.Left.Schema.Fields), len(node.OuterJoin.Right.Schema.Fields), leftKeyExprs, rightKeyExprs, node.OuterJoin.IsLeft, node.OuterJoin.IsRight), nil
+
+	case NodeTypeOrderSensitiveTransform:
+		source, err := node.OrderSensitiveTransform.Source.Materialize(ctx, env)
+		if err != nil {
+			return nil, fmt.Errorf("couldn't materialize order sensitive transform source: %w", err)
+		}
+		orderByKeyExprs := make([]execution.Expression, len(node.OrderSensitiveTransform.OrderByKey))
+		for i := range node.OrderSensitiveTransform.OrderByKey {
+			expr, err := node.OrderSensitiveTransform.OrderByKey[i].Materialize(ctx, env.WithRecordSchema(node.OrderSensitiveTransform.Source.Schema))
+			if err != nil {
+				return nil, fmt.Errorf("couldn't materialize order by key with index %d: %w", i, err)
+			}
+			orderByKeyExprs[i] = expr
+		}
+		var limit *execution.Expression
+		if node.OrderSensitiveTransform.Limit != nil {
+			expr, err := node.OrderSensitiveTransform.Limit.Materialize(ctx, env)
+			if err != nil {
+				return nil, fmt.Errorf("couldn't materialize limit expression: %w", err)
+			}
+			limit = &expr
+		}
+
+		if len(orderByKeyExprs) > 0 || (limit != nil && !node.OrderSensitiveTransform.Source.Schema.NoRetractions) {
+			return nodes.NewOrderSensitiveTransform(source, orderByKeyExprs, node.OrderSensitiveTransform.OrderByDirectionMultipliers, limit), nil
+		}
+
+		if limit != nil {
+			return nodes.NewLimit(source, *limit), nil
+		}
+
+		// Probably shouldn't happen...
+		return source, nil
 	}
 
 	panic(fmt.Sprintf("unexhaustive node type match: %d", node.NodeType))
